#!/usr/bin/env run-cargo-script
//! ```cargo
//! [package]
//! name = "rsconvssh"
//! version = "0.1.0"
//! authors = ["Vitaly _Vi Shukela <vi0oss@gmail.com>"]
//! edition = "2018"
//!
//! [dependencies]
//! serde_derive = "1.0.91"
//! serde = "1.0.91"
//! toml = "0.5.0"
//! encoding = "0.2.33"
//! log = "0.4.6"
//! simple-logging = "2.0.2"
//! log-derive = "0.3.0"
//! failure = "0.1.5"
//! ssh2 = "0.3.3"
//! base64 = "0.10.1"
//! ```

const CONFIG_FILE: &str = "config.toml";
const LOG_FILE: &str = "log.txt";

use encoding::types::{DecoderTrap, EncoderTrap};
use encoding::EncodingRef;
use failure::{bail, format_err};
use log::{debug, error, info, warn};
use log_derive::logfn;
use serde_derive::Deserialize;
use ssh2::Session;

type Result<T> = std::result::Result<T, failure::Error>;

#[derive(Deserialize, Debug)]
struct Server {
    host: String,
    port: u16,
    user: String,
    password: String,
    serverkey: Option<String>,
    runafterupload: Option<String>,
    debug: Option<bool>,
}

#[derive(Deserialize, Debug)]
struct Encodings {
    from: String,
    to: String,
}

#[derive(Deserialize, Debug)]
struct File {
    source: String,
    uploadpath: String,
    tmp: String,
}

#[derive(Deserialize, Debug)]
struct Config {
    upload: Server,
    encodings: Encodings,
    file: Vec<File>,
}

#[logfn(ok = "Trace", err = "Error")]
fn read_config() -> Result<Config> {
    let mut config = std::fs::read(CONFIG_FILE)?;
    let config : Vec<u8> = config.drain(..).filter(|x|*x!=0x0D).collect();
    Ok(toml::de::from_slice(config.as_ref())?)
}

#[logfn(err = "Error")]
fn get_encoding(s: &str) -> Result<EncodingRef> {
    Ok(encoding::label::encoding_from_whatwg_label(s)
        .ok_or_else(|| format_err!("Wrong encoding name"))?)
}

#[logfn(ok = "Trace", err = "Error")]
fn convert_file(src: &str, tmp: &str, from: EncodingRef, to: EncodingRef) -> Result<()> {
    let input = std::fs::read(src)?;
    let decoded = from
        .decode(input.as_ref(), DecoderTrap::Strict)
        .map_err(|_| format_err!("Failed to decode input file"))?;
    let encoded = to
        .encode(decoded.as_ref(), EncoderTrap::Strict)
        .map_err(|_| format_err!("Failed to encode to output file"))?;
    std::fs::write(tmp, encoded)?;
    Ok(())
}

#[logfn(err = "Error")]
fn connect_to_server(s: &Server) -> Result<std::net::TcpStream> {
    use std::net::ToSocketAddrs;

    for sa in (s.host.as_ref(), s.port).to_socket_addrs()? {
        match std::net::TcpStream::connect_timeout(&sa, std::time::Duration::from_secs(30)) {
            Err(e) => {
                warn!("Failed connection attempt to {}: {}", sa, e);
            }
            Ok(x) => return Ok(x),
        }
    }

    Err(format_err!("Failed to connect to server"))
}

#[logfn(err = "Error")]
fn ssh_auth(s: &Server, ssh: &mut Session, tcp: &std::net::TcpStream) -> Result<()> {
    ssh.set_compress(true);
    ssh.set_blocking(true);
    ssh.handshake(tcp)?;
    let server_key = ssh
        .host_key_hash(ssh2::HashType::Sha1)
        .ok_or_else(|| format_err!("Can't get SSH host key"))?;
    let server_key = base64::encode(server_key);

    if let Some(k) = s.serverkey.as_ref() {
        if &server_key != k {
            error!(
                "Invalid `serverkey` value. Expected: {}, Actual: {}",
                k, server_key
            );
            error!("This may indicate remote server reconfiguration or MitM hack attempt.");
            bail!("Invalid server key");
        }
    } else {
        warn!("`serverkey` parameter is not set. Set it to {}", server_key);
    }

    ssh.userauth_password(s.user.as_ref(), s.password.as_ref())?;

    Ok(())
}

#[logfn(ok = "Trace", err = "Error")]
fn upload_file(tmp: &str, uppath: &str, ssh: &mut Session) -> Result<()> {
    let data = std::fs::read(tmp)?;
    let mut remote_file =
        ssh.scp_send(std::path::Path::new(uppath), 0o644, data.len() as u64, None)?;
    debug!("scp_send completed");
    use std::io::Write;
    remote_file.write_all(data.as_ref())?;
    Ok(())
}

#[logfn(ok = "Trace", err = "Error")]
fn run_command(ssh: &mut Session, cmd: &str) -> Result<()> {
    let mut ch = ssh.channel_session()?;
    ch.exec(cmd)?;
    let mut stdout = String::new();
    let mut stderr = String::new();
    use std::io::Read;
    let _ = ch.read_to_string(&mut stdout)?;
    let _ = ch.stderr().read_to_string(&mut stderr)?;
    debug!("Remote command output: `{}` and `{}`", stdout, stderr);
    ch.wait_close()?;
    let es = ch.exit_status()?;
    if es != 0 {
        warn!(
            "Remote command exited with status {} and printed `{}` and `{}`",
            es, stdout, stderr,
        );
    }
    Ok(())
}

fn main() -> Result<()> {
    simple_logging::log_to_file(LOG_FILE, log::LevelFilter::Debug)?;

    let mut ssh = Session::new().ok_or_else(|| {
        error!("Internal error: failed to create SSH session");
        format_err!("Session::new failed")
    })?;

    let conf: Config = read_config()?;
    if conf.upload.debug != Some(true) {
        simple_logging::log_to_file(LOG_FILE, log::LevelFilter::Info)?;
    }
    info!("Configuration loaded");
    debug!("{:#?}", conf);

    let encoding_from = get_encoding(conf.encodings.from.as_ref())?;
    let encoding_to = get_encoding(conf.encodings.to.as_ref())?;

    info!(
        "Converting from {} to {}",
        encoding_from.name(),
        encoding_to.name()
    );

    for file in &conf.file {
        convert_file(
            file.source.as_ref(),
            file.tmp.as_ref(),
            encoding_from,
            encoding_to,
        )?;
        info!("Converted file {} to {}", file.source, file.tmp);
    }
    info!("Converted all files. Connecting to the server");

    let tcp = connect_to_server(&conf.upload)?;

    info!("Connected to server, logging in");

    ssh_auth(&conf.upload, &mut ssh, &tcp)?;

    info!("Logged in. Uploading files.");

    for file in conf.file {
        upload_file(file.tmp.as_ref(), file.uploadpath.as_ref(), &mut ssh)?;
        info!("Uploaded file {} to {}", file.tmp, file.uploadpath);
    }

    if let Some(r) = conf.upload.runafterupload {
        info!("Finished uploading. Running command `{}`", r);
        run_command(&mut ssh, r.as_ref())?;
        info!("Finished");
    } else {
        info!("Finished uploading. No command is set to run on server.");
    }

    ssh.disconnect(None, "Finished", None)?;

    Ok(())
}

// Config sample:
//
// [upload]
// host = "localhost"
// port = 22
// user =  "username"
// password = "passwd"
// serverkey = "..."
// runafterupload = "echo qwertyu"
// debug = false
//
// [encodings]
// from = "CP1251"
// to = "UTF-8"
//
// [[file]]
// source = "qqq_ru.txt"
// uploadpath = "/tmp/qqq.txt"
// tmp = "qqq.txt"
//
// [[file]]
// source = "eee_ru.txt"
// uploadpath = "/tmp/eee.txt"
// tmp = "eee.txt"
